# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2018 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import with_statement
import datetime
import logging
try:
    import simplejson as json
except ImportError:
    import json
import syntool_metadata.db

logger = logging.getLogger(__name__)
DAY_IN_SECONDS = 24 * 60 * 60


def parse_client_product_availability(cpa):
    """"""
    if cpa is None or len(cpa) == 0:
        return []

    cpa = json.loads(cpa)
    client_product_availability = []
    for i in xrange(0, len(cpa), 2):
        date_in_days = cpa[i]
        arity = cpa[i + 1]
        timestamp = date_in_days * DAY_IN_SECONDS
        date = datetime.datetime.utcfromtimestamp(timestamp)
        client_product_availability.append((arity, date))

    return client_product_availability


def get_availability(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    year = kwargs.get('year', None)
    if year is not None:
        if 'NaN' == year:
            year = None
        else:
            year = int(year)

    month = kwargs.get('month', None)
    if month is not None:
        if 'NaN' == month:
            month = None
        else:
            month = int(month)

    day = kwargs.get('day', None)
    if day is not None:
        if 'NaN' == day:
            day = None
        else:
            day = int(day)

    extent = kwargs.get('extent', None)
    if extent is not None and 1 > len(extent):
        extent = None

    products_filter = kwargs.get('products', None)
    if products_filter is not None:
        if 1 > len(products_filter):
            products_filter = None
        else:
            products_filter = products_filter.split(',')

    patterns = kwargs.get('patterns', None)
    if patterns is not None and 0 < len(patterns):
        patterns = dict(zip(products_filter, patterns.split(',')))
    else:
        patterns = None

    client_availability = kwargs.get('clientAvailability', None)
    if client_availability is not None and 0 < len(client_availability):
        client_availability = map(parse_client_product_availability,
                                  client_availability.split(';'))
        client_availability = dict(zip(products_filter, client_availability))
    else:
        client_availability = None

    storage_type = kwargs['webservice_cfg']['storage_type']
    with storage_type.get_session() as storage:
        result = syntool_metadata.db.get_availability(storage, year, month,
                                                      day, extent,
                                                      products_filter,
                                                      patterns,
                                                      client_availability)

    if result is None:
        result = {'years': {}, 'months': {}, 'days': {},
                  'error': ('No product available, please check that you did '
                            'not mispell product identifiers in the request')}

    if 'callback' in kwargs:
        return '{}({})'.format(kwargs['callback'], json.dumps(result))
    else:
        return result


def get_datasets(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'minDate', 'maxDate',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    extent = kwargs.get('extent', None)
    if extent is not None and 1 > len(extent):
        extent = None

    products_filter = kwargs.get('products', None)
    if products_filter is not None:
        if 1 > len(products_filter):
            products_filter = None
        else:
            products_filter = products_filter.split(',')

    patterns = kwargs.get('patterns', None)
    if patterns is not None and 0 < len(patterns):
        patterns = dict(zip(products_filter, patterns.split(',')))
    else:
        patterns = None

    min_date = datetime.datetime.utcfromtimestamp(float(kwargs['minDate']))
    max_date = datetime.datetime.utcfromtimestamp(float(kwargs['maxDate']))

    storage_type = kwargs['webservice_cfg']['storage_type']
    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    results_limit = kwargs['webservice_cfg']['results_limit']
    with_features = kwargs['webservice_cfg']['with_features']
    with storage_type.get_session() as storage:
        datasets = syntool_metadata.db.get_datasets(storage, 'data/ingested',
                                                    ingested_dir, min_date,
                                                    max_date, extent,
                                                    products_filter, patterns,
                                                    results_limit,
                                                    with_features)

    result = {'events': datasets}

    if 'callback' in kwargs:
        return '{}({})'.format(kwargs['callback'], json.dumps(result))
    else:
        return result


def find_nearest_dataset(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'date',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    extent = kwargs.get('extent', None)
    if extent is not None and 1 > len(extent):
        extent = None

    products_filter = kwargs.get('products', None)
    if products_filter is not None:
        if 1 > len(products_filter):
            products_filter = None
        else:
            products_filter = products_filter.split(',')

    patterns = kwargs.get('patterns', None)
    if patterns is not None and 0 < len(patterns):
        patterns = dict(zip(products_filter, patterns.split(',')))
    else:
        patterns = None

    direction = kwargs.get('direction', None)
    if direction is not None and 1 > len(direction):
        direction = None

    ref_datetime = datetime.datetime.utcfromtimestamp(float(kwargs['date']))

    time_step = kwargs.get('time_step', 300)
    try:
        time_step = int(time_step)
    except ValueError:
        msg = 'time_step must be an integer value'
        return {'ok': False, 'details': msg}

    storage_type = kwargs['webservice_cfg']['storage_type']
    with storage_type.get_session() as storage:
        ts, before, after = syntool_metadata.db.find_nearest(storage,
                                                             direction,
                                                             ref_datetime,
                                                             extent,
                                                             products_filter,
                                                             patterns,
                                                             time_step)

    result = {'nearestDataDate': ts,
              'before_delta': before,
              'after_delta': after}

    if 'callback' in kwargs:
        return '{}({})'.format(kwargs['callback'], json.dumps(result))
    else:
        return result


def list_products(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    storage_type = kwargs['webservice_cfg']['storage_type']
    with storage_type.get_session() as storage:
        results = syntool_metadata.db.list_products(storage)
    return '\n'.join(['{} : {}'.format(k, v) for k, v in results.iteritems()])


def get_granule_info(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'product_id', 'dataset_name')
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    result = syntool_metadata.db.get_granule_info(ingested_dir,
                                                  kwargs['product_id'],
                                                  kwargs['dataset_name'])
    return json.dumps(result)
