# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2018 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sqlalchemy.ext.declarative
from sqlalchemy import func, desc
from sqlalchemy.types import UserDefinedType
from sqlalchemy.engine import Engine, reflection
import ConfigParser
import contextlib
import datetime
import calendar
import logging
import os

logger = logging.getLogger(__name__)
Base = sqlalchemy.ext.declarative.declarative_base()


class UnknownGlobalExtent(Exception):
    def __init__(self, product_id):
        self.product_id = product_id


class MissingProductsTable(Exception):
    pass


# Mappings
# -----------------------------------------------------------------------------
class Geometry(UserDefinedType):
    def get_col_spec(self):
        return "GEOMETRY"

    def bind_expression(self, bindvalue):
        return func.GeomFromText(bindvalue, type_=self)

    def column_expression(self, col):
        return func.AsText(col, type_=self)


class Product(Base):
    """ """
    __tablename__ = 'products'
    product_id = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                   primary_key=True)
    shortname = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                  nullable=False)
    type = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255), nullable=False)

    def __init__(self):
        """ """
        self.type = 'ZXY'


class Dataset():
    """ """
    dataset_name = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                     nullable=False, primary_key=True)
    relative_path = sqlalchemy.Column(sqlalchemy.types.VARCHAR(255),
                                      nullable=False)
    begin_datetime = sqlalchemy.Column(sqlalchemy.types.DATETIME,
                                       nullable=False)
    end_datetime = sqlalchemy.Column(sqlalchemy.types.DATETIME, nullable=False)
    min_zoom_level = sqlalchemy.Column(sqlalchemy.types.INTEGER,
                                       nullable=False)
    max_zoom_level = sqlalchemy.Column(sqlalchemy.types.INTEGER,
                                       nullable=False)
    resolutions = sqlalchemy.Column(sqlalchemy.types.TEXT, nullable=False)
    bbox_text = sqlalchemy.Column(sqlalchemy.types.TEXT, nullable=False)
    bbox_geometry = sqlalchemy.Column(Geometry, nullable=True)
    shape_text = sqlalchemy.Column(sqlalchemy.types.TEXT, nullable=False)
    shape_geometry = sqlalchemy.Column(Geometry, nullable=False)


# Storage implementation
# -----------------------------------------------------------------------------
@sqlalchemy.event.listens_for(Engine, "connect")
def set_mysql_time_zone(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("SET time_zone = '+0:0'")
    cursor.close()


class Storage(object):
    """ """

    def __init__(self, db_url, *args, **kwargs):
        """ """
        super(Storage, self).__init__(*args, **kwargs)
        self.product_tables = {}
        self.__engine = sqlalchemy.create_engine(db_url, echo=False,
                                                 pool_recycle=10)
        self.Base = sqlalchemy.ext.declarative.declarative_base()

    def list_tables(self):
        """"""
        insp = reflection.Inspector.from_engine(self.__engine)
        existing_tables = insp.get_table_names()
        return existing_tables

    def get_product_table(self, product_id):
        fixed_product_id = product_id.replace(' ', '_')
        if fixed_product_id not in self.product_tables:
            table_name = 'product_{}'.format(fixed_product_id)
            product_table = type('Dataset_{}'.format(fixed_product_id),
                                 (self.Base, Dataset),
                                 {'__tablename__': table_name})
            self.product_tables[fixed_product_id] = product_table
        return self.product_tables[fixed_product_id]

    def query(self, *args, **kwargs):
        """"""
        return self.__session.query(*args, **kwargs)

    @contextlib.contextmanager
    def get_session(self):
        """ """
        Session = sqlalchemy.orm.sessionmaker(bind=self.__engine)
        self.__session = Session()
        try:
            yield self
            self.__session.commit()
        except Exception:
            self.__session.rollback()
            raise
        finally:
            self.__session.close()
            logger.info('Session closed')


def get_product_table_name(product_id):
    return 'product_{}'.format(product_id.replace(' ', '_'))


def build_query(q, dataset_cls, bbox_text, pattern):
    """"""
    bbox_polygon = None
    if bbox_text is not None:
        west, south, east, north = bbox_text.split(',')
        bbox_polygon = 'POLYGON(({} {}, {} {}, {} {}, {} {}, {} {}))'
        bbox_polygon = bbox_polygon.format(west, north, east, north, east,
                                           south, west, south, west, north)
        geom_intersect = func.Intersects(dataset_cls.shape_geometry,
                                         func.GeomFromText(bbox_polygon))
        q = q.filter(sqlalchemy.or_(dataset_cls.shape_text == 'POINT(0 0)',
                                    geom_intersect))

    if pattern is not None:
        sql_pattern = '%{}%'.format(pattern)
        q = q.filter(dataset_cls.dataset_name.like(sql_pattern))

    return q


def get_image_info(storage, product_id, dataset_id):
    """ """
    dataset_cls = storage.get_product_table(product_id)
    q = storage.query(dataset_cls.bbox_text)
    q = q.filter(dataset_cls.dataset_name == dataset_id)
    q = q.limit(1)
    result = q.first()
    if result is None:
        return None
    if 'POINT(0 0)' == result[0]:
        is_webmercator = (product_id.startswith('900913_') or
                          product_id.startswith('3857_'))
        if is_webmercator is True:
            west = -20037508.34
            north = 20037508.34
            east = 20037508.34
            south = -20037508.34
        elif product_id.startswith('3413_'):
            west = -5000000
            north = 5000000
            east = 5000000
            south = -5000000
        else:
            raise UnknownGlobalExtent(product_id)
    else:
        point0, _, point2, _, _ = result[0][9:-2].split(',')
        west, north = map(float, point0.split(' '))
        east, south = map(float, point2.split(' '))
    return west, north, east, south


def update_availability(result, tkey, ttype, arity, product_index):
    """ """
    if tkey not in result[ttype]:
        result[ttype][tkey] = {'datasets': arity,
                               'products': [product_index]}
    else:
        result[ttype][tkey]['datasets'] += arity
    if product_index not in result[ttype][tkey]['products']:
        result[ttype][tkey]['products'].append(product_index)


def get_availability(storage, year, month, day, bbox_text, requested_products,
                     patterns, client_availability):
    """ """
    existing_tables = storage.list_tables()
    products = [p for p in requested_products
                if get_product_table_name(p) in existing_tables]
    if client_availability is not None:
        client_products = [p for p in requested_products
                           if p in client_availability and p not in products]
        products.extend(client_products)
    if 0 >= len(products):
        return None

    # Temporal filter
    if year is None:
        date_start = None
        date_stop = None
    elif month is None:
        date_start = datetime.date(year, 1, 1)
        days_extent = 365
        if calendar.isleap(year):
            days_extent += 1
        date_stop = date_start + datetime.timedelta(days=days_extent)
    elif day is None:
        date_start = datetime.date(year, month, 1)
        _, days_extent = calendar.monthrange(year, month)
        date_stop = date_start + datetime.timedelta(days=days_extent)
    else:
        date_start = datetime.date(year, month, day)
        date_stop = date_start + datetime.timedelta(days=1)

    result = {'years': {}, 'months': {}, 'days': {}}

    all_days = []
    product_index = 0
    for product_id in products:
        product_index = product_index + 1

        pattern = None
        if patterns is not None:
            pattern = patterns.get(product_id, None)

        cproduct_availability = []
        if client_availability is not None:
            cproduct_availability = client_availability.get(product_id, [])

        product_availability = get_product_availability(storage, product_id,
                                                        date_start, date_stop,
                                                        bbox_text, pattern)
        product_all_days = []

        # Include availability for products defined on the server side
        for arity, start, stop in product_availability:
            dt = start
            while dt <= stop:
                dt_availability = (dt.year, dt.month, dt.day)
                product_all_days.append(dt_availability)
                update_availability(result, dt.year, 'years', arity,
                                    product_index)
                if dt.year == year:
                    update_availability(result, dt.month, 'months', arity,
                                        product_index)
                    if dt.month == month:
                        update_availability(result, dt.day, 'days', arity,
                                            product_index)

                dt = dt + datetime.timedelta(days=1)

        # Include availability for products defined on the client side
        for arity, start in cproduct_availability:
            product_all_days.append((start.year, start.month, start.day))
            update_availability(result, start.year, 'years', arity,
                                product_index)
            if start.year == year:
                update_availability(result, start.month, 'months', arity,
                                    product_index)
                if start.month == month:
                    update_availability(result, start.day, 'days', arity,
                                        product_index)

        all_days.append(set(product_all_days))

    coloc_days = set.intersection(*all_days)
    coloc_years = set([x[0] for x in coloc_days])
    coloc_months = set([x[1] for x in coloc_days if x[0] == year])

    for y in result['years']:
        if y in coloc_years:
            result['years'][y]['coloc'] = products
        else:
            result['years'][y]['coloc'] = []
    for m in result['months']:
        if m in coloc_months:
            result['months'][m]['coloc'] = products
        else:
            result['months'][m]['coloc'] = []

    return result


def get_product_availability(storage, product_id, date_start, date_stop,
                             bbox_text, pattern):
    """ """
    """
    sqlQuery = "SELECT COUNT(*) AS nbDatasets, DATE_FORMAT(begin_datetime,\
                '%Y/%m/%d') AS begin_datetime2, " \
                "DATE_FORMAT(end_datetime, '%Y/%m/%d') AS end_datetime2 \
                FROM `product_" + productId.replace(' ', '_') + "` " \
                "GROUP BY begin_datetime2, end_datetime2"
    """
    existing_tables = storage.list_tables()
    table_name = get_product_table_name(product_id)
    if table_name not in existing_tables:
        return []
    dataset_cls = storage.get_product_table(product_id)
    nb_datasets = func.count(dataset_cls.dataset_name).label('nb_datasets')
    dataset_begin = sqlalchemy.func.date(dataset_cls.begin_datetime)
    dataset_end = sqlalchemy.func.date(dataset_cls.end_datetime)
    q = storage.query(nb_datasets, dataset_begin, dataset_end)
    q = build_query(q, dataset_cls, bbox_text, pattern)
    q = q.group_by(dataset_begin, dataset_end)

    return q.all()


def find_nearest(storage, direction, current, bbox_text, requested_products,
                 patterns, search_step=60):
    """ """
    existing_tables = storage.list_tables()
    products = [p for p in requested_products
                if get_product_table_name(p) in existing_tables]

    after_delta = float('inf')
    before_delta = float('inf')
    for product_id in products:
        pattern = None
        if patterns is not None:
            pattern = patterns.get(product_id, None)

        dataset_cls = storage.get_product_table(product_id)

        mid_ts = ((func.unix_timestamp(dataset_cls.begin_datetime) +
                   func.unix_timestamp(dataset_cls.end_datetime)) / 2)
        mid_ts = mid_ts.label('mid_ts')
        current_ts = calendar.timegm(current.timetuple())

        # Look for datasets in the future
        if direction in ['nearest', 'after', None]:
            q = storage.query(mid_ts)
            q = q.filter(mid_ts > current_ts + search_step)
            q = build_query(q, dataset_cls, bbox_text, pattern)
            q = q.order_by(mid_ts)
            q = q.limit(1)
            result = q.first()
            if result is not None:
                delta = int(result[0]) - current_ts
                if delta < after_delta:
                    after_delta = delta

        # Look for datasets in the past
        if direction in ['nearest', 'before', None]:
            q = storage.query(mid_ts)
            q = q.filter(mid_ts < current_ts - search_step)
            q = build_query(q, dataset_cls, bbox_text, pattern)
            q = q.order_by(desc(mid_ts))
            q = q.limit(1)
            result = q.first()
            if result is not None:
                delta = current_ts - int(result[0])
                if delta < before_delta:
                    before_delta = delta

    delta = min(after_delta, before_delta)
    if delta == float('inf'):
        new_ts = current_ts
    elif delta == after_delta:
        new_ts = current_ts + delta
    else:
        new_ts = current_ts - delta

    if before_delta == float('inf'):
        before_delta = -1
    if after_delta == float('inf'):
        after_delta = -1

    return new_ts, before_delta, after_delta


def format_dataset(product_id, product_type, data_url, data_path, dataset,
                   with_features):
    """"""
    resolutions = []

    # Remove suffix for DB entries used to bypass cross-IDL issues
    dataset_name = dataset.dataset_name
    if dataset_name.endswith('_XIDLfix'):
        dataset_name = dataset_name[:-8]

    dataset_uri = '{}/{}/{}/'.format(data_url, product_id, dataset_name)

    if 'POINT(0 0)' == dataset.bbox_text:
        is_webmercator = (product_id.startswith('900913_') or
                          product_id.startswith('3857_'))
        if is_webmercator is True:
            west = -20037508.34
            north = 20037508.34
            east = 20037508.34
            south = -20037508.34
        elif product_id.startswith('3413_'):
            west = -5000000
            north = 5000000
            east = 5000000
            south = -5000000
        else:
            msg = 'Global extent unknown for {}'.format(product_id)
            raise Exception(msg)
    else:
        point0, _, point2, _, _ = dataset.bbox_text[9:-2].split(',')
        west, north = map(float, point0.split(' '))
        east, south = map(float, point2.split(' '))

    if 'IMAGE' == product_type:
        resolutions = dataset.resolutions
        uri = '{}imageLayer.png'.format(dataset_uri)
    elif product_type in ['VECTOR_FIELD', 'STREAMLINES']:
        resolutions = dataset.resolutions
        uri = '{}vectorFieldLayer.png'.format(dataset_uri)
    elif 'ZXY' == product_type:
        resolutions = map(lambda x: x.split(':')[1],
                          dataset.resolutions.split(','))
        uri = '{}tiles.zxy/'.format(dataset_uri)
    elif 'TMS' == product_type:
        resolutions = map(lambda x: x.split(':')[1],
                          dataset.resolutions.split(','))
        uri = '{}tiles/'.format(dataset_uri)
    else:
        uri = dataset_uri

    return {'productId': product_id,
            'datasetId': '{}-{}'.format(product_id, dataset_name),
            'title': dataset_name,
            'uri': uri,
            'start': dataset.begin_datetime.strftime('%Y-%m-%dT%H:%M:%S'),
            'end': dataset.end_datetime.strftime('%Y-%m-%dT%H:%M:%S'),
            'mapMinZoom': "{}".format(dataset.min_zoom_level),
            'mapMaxZoom': "{}".format(dataset.max_zoom_level),
            'mapBounds': {'north': '{}'.format(north),
                          'south': '{}'.format(south),
                          'west': '{}'.format(west),
                          'east': '{}'.format(east)},
            'polygon': dataset.shape_text,
            'point': {'lon': 0.5 * (east + west),
                      'lat': 0.5 * (north + south)},
            'resolutions': resolutions,
            'type': product_type}


def get_datasets(storage, data_url, data_path, start, stop, bbox_text,
                 requested_products, patterns, results_limit, with_features):
    """ """
    existing_tables = storage.list_tables()
    products = [p for p in requested_products
                if get_product_table_name(p) in existing_tables]

    result = []
    bucket_size = results_limit
    for product_id in products:
        if 0 >= bucket_size:
            break

        q0 = storage.query(Product.type)
        q0 = q0.filter(Product.product_id == product_id)
        product_type = q0.first()[0]

        pattern = None
        if patterns is not None:
            pattern = patterns.get(product_id, None)

        dataset_cls = storage.get_product_table(product_id)
        q1 = storage.query(dataset_cls.dataset_name)
        q1 = q1.filter(dataset_cls.begin_datetime < stop)
        q1 = q1.filter(dataset_cls.end_datetime > start)

        q = storage.query(dataset_cls.dataset_name, dataset_cls.resolutions,
                          dataset_cls.begin_datetime, dataset_cls.end_datetime,
                          dataset_cls.min_zoom_level,
                          dataset_cls.max_zoom_level,
                          dataset_cls.bbox_text, dataset_cls.shape_text)
        q = q.filter(dataset_cls.dataset_name.in_(q1))
        q = build_query(q, dataset_cls, bbox_text, pattern)
        q = q.order_by(dataset_cls.begin_datetime)
        q = q.limit(bucket_size)
        product_datasets = [format_dataset(product_id, product_type, data_url,
                                           data_path, x, with_features)
                            for x in q.all()]
        bucket_size = bucket_size - len(product_datasets)
        result.extend(product_datasets)
    return result


def list_products(storage, **kwargs):
    existing_tables = storage.list_tables()
    if 'products' not in existing_tables:
        raise MissingProductsTable()
    q = storage.query(Product.product_id, Product.shortname)
    results = {k: v for k, v in q.all()}
    return results


def search_datasets(storage, start, stop, bbox_text, requested_products,
                    patterns):
    """ """
    existing_tables = storage.list_tables()
    products = [p for p in requested_products
                if get_product_table_name(p) in existing_tables]

    # Spatial filter
    bbox_polygon = None
    if bbox_text is not None:
        west, south, east, north = bbox_text.split(',')
        bbox_polygon = 'POLYGON(({} {}, {} {}, {} {}, {} {}, {} {}))'
        bbox_polygon = bbox_polygon.format(west, north, east, north, east,
                                           south, west, south, west, north)

    results = {}

    for product_id in products:
        pattern = None
        if patterns is not None:
            pattern = patterns.get(product_id, None)

        dataset_cls = storage.get_product_table(product_id)
        q = storage.query(dataset_cls.dataset_name)
        q = build_query(q, dataset_cls, bbox_text, pattern)
        q = q.filter(dataset_cls.begin_datetime < stop)
        q = q.filter(dataset_cls.end_datetime > start)
        q = q.order_by(dataset_cls.begin_datetime)
        results[product_id] = [x[0] for x in q.all()]
    return results


def get_granule_info(data_path, product_id, dataset_name):
    """"""
    features = []
    features_dir = os.path.join(data_path, product_id, dataset_name,
                                'features')
    if os.path.isdir(features_dir):
        for root, dirs, files in os.walk(features_dir):
            ini_features = [os.path.join(root, fn) for fn in files
                            if fn.lower().endswith('.ini')]
            for ini_path in ini_features:
                ini_parser = ConfigParser.ConfigParser()
                ini_parser.read(ini_path)
                feature = dict(ini_parser._sections)
                for k in feature:
                    feature[k].pop('__name__', None)
                features.append(feature)
    return features
