# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2018 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
import datetime
from webob import Response
import syntool_metadata.histogram
import syntool_metadata.extract
try:
    import simplejson as json
except ImportError:
    import json

logger = logging.getLogger(__name__)


def search_datasets(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'minDate', 'maxDate', 'products', 'format',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    output_format = kwargs.get('format', 'url')
    if output_format not in ['url', 'path']:
        msg = ('Output format "{}" is not supported, use either "url" or '
               '"path"')
        return {'ok': False, 'details': msg}

    if 'url' == output_format and 'SERVER_HOST' not in environ:
        msg = ('Webserver misconfigured: missing SERVER_HOST environment '
               'variable')
        return {'ok': False, 'details': msg}

    extent = kwargs.get('extent', None)
    if extent is not None and 1 > len(extent):
        extent = None

    products_filter = kwargs.get('products', None)
    if products_filter is not None:
        if 1 > len(products_filter):
            products_filter = None
        else:
            products_filter = products_filter.split(',')

    patterns = kwargs.get('patterns', None)
    if patterns is not None and 0 < len(patterns):
        patterns = dict(zip(products_filter, patterns.split(',')))
    else:
        patterns = None

    min_date = datetime.datetime.utcfromtimestamp(float(kwargs['minDate']))
    max_date = datetime.datetime.utcfromtimestamp(float(kwargs['maxDate']))

    storage_type = kwargs['webservice_cfg']['storage_type']
    with storage_type.get_session() as storage:
        datasets = storage.search_datasets(min_date,
                                           max_date,
                                           extent,
                                           products_filter,
                                           patterns)

    result = []
    if 'url' == output_format:
        server_host = environ['SERVER_HOST']
        if not server_host.endswith('/'):
            server_host = '{}/'.format(server_host)
        for product_id, product_items in datasets.iteritems():
            product_url = '{}data/ingested/{}'.format(server_host, product_id)
            product_datasets = map(lambda x: '{}/{}'.format(product_url, x),
                                   product_items)
            result.extend(product_datasets)
    elif 'path' == output_format:
        for product_id, product_items in datasets.iteritems():
            product_datasets = map(lambda x: os.path.join(product_id, x),
                                   product_items)
            result.extend(product_datasets)

    return '\n'.join(result)


def histogram(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'productId', 'datasetId', 'x1', 'y1', 'x2',
                'y2', 'zoomLevel',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    dataset_id, dataset_type = kwargs['datasetId'].split('+', 1)

    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    result = syntool_metadata.histogram.get(ingested_dir,
                                            kwargs['productId'],
                                            dataset_id,
                                            dataset_type,
                                            float(kwargs['x1']),
                                            float(kwargs['y1']),
                                            float(kwargs['x2']),
                                            float(kwargs['y2']),
                                            int(kwargs['zoomLevel']))

    if 'callback' in kwargs:
        return '{}({})'.format(kwargs['callback'], json.dumps(result))
    else:
        return result


def download_png(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'datasets', 'x1', 'y1', 'x2', 'y2',
                'resolution',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    if 'SERVER_HOST' not in environ:
        msg = ('Webserver misconfigured: missing SERVER_HOST environment '
               'variable')
        return {'ok': False, 'details': msg}

    storage_type = kwargs['webservice_cfg']['storage_type']
    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    download_dir = kwargs['webservice_cfg']['download_dir']
    with storage_type.get_session() as storage:
        results = syntool_metadata.extract.get(ingested_dir,
                                               kwargs['datasets'].split(','),
                                               float(kwargs['x1']),
                                               float(kwargs['y1']),
                                               float(kwargs['x2']),
                                               float(kwargs['y2']),
                                               float(kwargs['resolution']),
                                               'PNG',
                                               environ['SERVER_HOST'],
                                               download_dir,
                                               storage)
    with open(results, 'r') as f:
        content = f.read()

    n = datetime.datetime.now()
    now_str = n.strftime('%Y%m%d_%H%M%S')
    default_filename = 'syntool-archive-{}.tgz'.format(now_str)
    r = Response()
    r.body = content
    r.content_type = 'application/octet-stream'
    r.content_disposition = 'attachement; filename={}'.format(default_filename)
    r.content_description = 'File transfer'
    return r


def download_numpy(environ, *args, **kwargs):
    """ """
    required = ('webservice_cfg', 'datasets', 'x1', 'y1', 'x2', 'y2',
                'resolution',)
    missing = [x for x in required if x not in kwargs or 0 >= len(x)]
    if 0 < len(missing):
        msg = '{}  parameter(s) must be present in the request'
        msg = msg.format(', '.join(missing))
        return {'ok': False, 'details': msg}

    if 'SERVER_HOST' not in environ:
        msg = ('Webserver misconfigured: missing SERVER_HOST environment '
               'variable')
        return {'ok': False, 'details': msg}

    storage_type = kwargs['webservice_cfg']['storage_type']
    ingested_dir = kwargs['webservice_cfg']['ingested_dir']
    download_dir = kwargs['webservice_cfg']['download_dir']
    with storage_type.get_session() as storage:
        results = syntool_metadata.extract.get(ingested_dir,
                                               kwargs['datasets'].split(','),
                                               float(kwargs['x1']),
                                               float(kwargs['y1']),
                                               float(kwargs['x2']),
                                               float(kwargs['y2']),
                                               float(kwargs['resolution']),
                                               'NUMPY',
                                               environ['SERVER_HOST'],
                                               download_dir,
                                               storage)

    if 'callback' in kwargs:
        return '{}({})'.format(kwargs['callback'], json.dumps(results))
    else:
        return results
