# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2018 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
import ConfigParser
import syntool_metadata.db
from routr import route, POST, GET

logger = logging.getLogger(__name__)


class SettingsFileEnvMissing(Exception):
    """"""
    pass


class SettingsFileNotFound(Exception):
    """"""
    def __init__(self, path):
        self.path = path


def get_ingested_dir(root_dir):
    """"""
    data_dir = os.path.join(root_dir, 'data')
    root_ingested_path = os.path.join(root_dir, 'ingested')
    root_data_ingested_path = os.path.join(data_dir, 'ingested')
    ingested_dir = '/dev/null'
    if os.path.isdir(root_ingested_path):
        ingested_dir = root_ingested_path
    elif os.path.isdir(root_data_ingested_path):
        ingested_dir = root_data_ingested_path
    return ingested_dir


def load_cfg():
    """"""
    settings_ini_path = os.environ.get('SYNTOOL_INI', None)
    if settings_ini_path is None:
        raise SettingsFileEnvMissing()
    if not os.path.exists(settings_ini_path):
        raise SettingsFileNotFound(settings_ini_path)

    ini_parser = ConfigParser.ConfigParser()
    ini_parser.read(settings_ini_path)

    db_cfg = ini_parser._sections['database']
    db_uri = 'mysql://{}:{}@{}:{}/{}'.format(db_cfg.get('user', 'syntool'),
                                             db_cfg.get('password', 'syntool'),
                                             db_cfg.get('host', 'localhost'),
                                             db_cfg.get('port', 3306),
                                             db_cfg.get('name', 'syntool'))
    storage_type = syntool_metadata.db.Storage(db_uri)

    root_dir = ini_parser._sections['general']['root_dir']
    download_dir = ini_parser._sections['general']['download_dir']
    results_limit = int(ini_parser._sections['database'].get('results_limit',
                                                             5000))
    full_service = False
    if ini_parser.has_option('general', 'full-service'):
        full_service = ini_parser.getboolean('general', 'full-service')

    with_features = True
    if ini_parser.has_option('general', 'with-features'):
        with_features = ini_parser.getboolean('general', 'with-features')

    url_prefix = os.environ.get('URL_PREFIX', '')
    tile_servers = os.environ.get('SYNTOOL_TILE_SERVERS', '').split(',')
    tile_servers = filter(lambda x: 0 < len(x), tile_servers)

    ingested_dir = get_ingested_dir(root_dir)

    webservice_cfg = {'storage_type': storage_type,
                      'root_dir': root_dir,
                      'ingested_dir': ingested_dir,
                      'download_dir': download_dir,
                      'results_limit': results_limit,
                      'full_service': full_service,
                      'url_prefix': url_prefix,
                      'tile_servers': tile_servers,
                      'with_features': with_features}
    return webservice_cfg


def init_routes(url_prefix, full_service):
    """"""
    import syntool_metadata.basic
    basic_routes = route(route(GET, '/availability.service.php',
                               syntool_metadata.basic.get_availability,
                               name='get_availability'),
                         route(POST, '/availability.service.php',
                               syntool_metadata.basic.get_availability,
                               name='post_availability'),
                         route(GET, '/data-noRegion.service.php',
                               syntool_metadata.basic.get_datasets,
                               name='get_datasets'),
                         route(GET, '/findnearestdata.service.php',
                               syntool_metadata.basic.find_nearest_dataset,
                               name='get_nearest_dataset'),
                         route(GET, '/listProducts/',
                               syntool_metadata.basic.list_products,
                               name='get_products_list'),
                         route(GET, '/getGranuleInfo/',
                               syntool_metadata.basic.get_granule_info,
                               name='get_granule_info'))

    routes = None
    if full_service is True:
        import syntool_metadata.extra
        extra_routes = route(route(GET, '/histogram.service.php',
                                   syntool_metadata.extra.histogram,
                                   name='get_histogram'),
                             route(GET, 'downloadRaster.service.php',
                                   syntool_metadata.extra.download_png,
                                   name='get_raster'),
                             route(GET, 'downloadNumpy.service.php',
                                   syntool_metadata.extra.download_numpy,
                                   name='get_numpy'),
                             route(GET, 'search',
                                   syntool_metadata.extra.search_datasets,
                                   name='get_search'))

        routes = route(url_prefix, basic_routes, extra_routes)
    else:
        routes = route(url_prefix, basic_routes)

    return routes
